--Aufgabe 6
--Beispiel 1

--Notwendige Definitionen
data Stack a = Stk [a] | NoStk deriving (Eq,Show)
--data Maybe a = Just a | Nothing deriving (Eq,Show)

--parameterlose Funktion liefert leeren Stack zurück
empty :: (Eq a,Show a) => Stack a
empty = Stk []

--überprüft, ob übergebener Stack der leere Stack ist
isEmpty :: (Eq a,Show a) => Stack a -> Bool
isEmpty (Stk []) = True
isEmpty _ = False

--zeigt das oberste Element des Stacks an
top1 :: (Eq a,Show a) => Stack a -> a
top1 (Stk (x:xs)) = x
top1 _ = error "Invalid Argument"

--zeigt das oberste Element des Stacks an
top2 :: (Eq a,Show a) => Stack a -> Maybe a
top2 (Stk(x:xs)) = Just x
top2 _ = Nothing

--entfernt das oberste Element des Stacks
pop :: (Eq a,Show a) => Stack a -> Stack a
pop (Stk []) = NoStk
pop (NoStk) = NoStk
pop (Stk (a:as)) = Stk as 

--fügt ein neues Elemtent an der obersten Stelle im Stack ein
push :: (Eq a,Show a) => a -> Stack a -> Stack a
push e (NoStk) = NoStk
push e (Stk x) = Stk new where
	new = e:x

--Beispiel 2
--Rechenvorschriften für Numeral aus Aufgabe 3

canonize :: Numeral -> Numeral
canonize (Num (s, d))
 | d == [] = error "Invalid Argument"
 | (head d == Zero) && (tail d == []) = Num (Pos,d)
 | otherwise = Num (s, (canonizeH d))

--canonizeH nimmt Digits und entfernt solange führende Zero, bis One oder Two "führt"

canonizeH :: Digits -> Digits
canonizeH d
 | (head d == Zero) && (tail d == []) = d
 | head d == Zero = canonizeH (drop 1 d)
 | otherwise = d

--int2num zur Umwandlung von Dezimalzahlen zu Num, ruft canonize, int2numH auf
--Anmerkung: Digits für Pos/Neg Nums sind gleich, nur das Vorzeichen muss geändert werden

int2num :: Integer -> Numeral
int2num n 
 | n == 0 = Num (Pos, [Zero])
 | n < 0 = Num (Neg, (int2numH ((-1) * n) []))
 | otherwise = canonize (Num (Pos, (int2numH n [])))

--Da Sonderfall 0 schon in int2num abgedeckt, ist n == 0 Abbruchbedingung (sonst wäre der Rückgabewert ungültig)
--Wenn die Zahl durch 3 teilbar, füge vorne 0 an
--Wenn die Zahl mod 3 == 1, füge 1 vorne an
--Andernfalls füge 2 an
--Nächste Rekursion mit div n 3

int2numH :: Integer -> Digits -> Digits
int2numH n d
 | n == 0 = d
 | mod n 3 == 0 = int2numH (div n 3) (Zero : d)
 | mod n 3 == 1 = int2numH (div n 3) (One : d)
 | otherwise = int2numH (div n 3) (Two : d)

--num2int wandelt Numerale in Dezimalzahlen um, "leere" Zahlen werden nicht akzeptiert
--ruft canonize, num2intH auf

num2int :: Numeral -> Integer
num2int (Num (s,d))
 | d == [] = error "Invalid Argument"
 | s == Neg = (-1) * (num2intH d)
 | otherwise = num2intH (canonizeH d)

--Wenn letzte Stelle Zero, addiere 0 
--Wenn letzte Stelle One, addiere 1
--Wenn letzte Stelle Two, addiere 2
--"Restzahl" wird mit 3 für Stelligkeit multipliziert

num2intH :: Digits -> Integer
num2intH d
 | d == [] = 0
 | last d == Zero = (3 * (num2intH(init d))) + 0
 | last d == One = (3 * (num2intH(init d))) + 1
 | otherwise = (3 * (num2intH(init d))) + 2

--Notwendige Definitionen

data Digit      = Zero | One | Two deriving (Eq,Enum,Show)
type Digits     = [Digit]
data Sign       = Pos | Neg  deriving (Eq,Show)

newtype Numeral = Num (Sign,Digits) deriving (Eq,Show)

data Operator   = Plus | Times | Minus deriving (Eq,Show)

data Variable   = A | B | C deriving (Eq,Show)
data Expression = Cst Numeral
 | Var Variable
 | Exp Expression Expression Operator deriving (Eq,Show)
type State      = Variable -> Numeral

--Funktion eval wertet den Ausdruck rekursiv aus
eval :: Expression -> State -> Integer
eval (Cst num) s = num2int num
eval (Var v) s = num2int(s v) 
eval (Exp x y op) s = operate (eval x s) (eval y s) op  where
	operate a b Plus = a + b
	operate a b Times = a * b
	operate a b Minus = a - b

--Beispiel 3
--Notwendige Definitionen
data CVO    = Cop Numeral -- Konstantenoperand
 | Vop Char -- Variablenoperand
 | OpPlus   -- Operator fuer Addition
 | OpTimes  -- Operator fuer Multiplikation
 | OpMinus  -- Operator fuer Subtraktion
 deriving (Eq,Show)
type Expr   = [CVO]
type State2 = Char -> Numeral

s2 :: State2 
s2 'A' = Num (Pos,[One])         -- entspricht s2(’A’) = 0
s2 'B' = Num (Neg,[One,Zero,One]) -- s2(’B’) = -10
s2 'C' = Num (Pos,[One,One,One])  -- s2(’C’) = 13
s2 _   = Num (Pos,[One])

e = [Vop 'A',Vop 'B',OpMinus,Vop 'C',OpPlus]

constant :: CVO -> Integer
constant (Cop c) = num2int c

operating :: Integer -> Integer -> CVO -> Integer
operating b a OpTimes = a * b
operating b a OpPlus = a + b
operating b a OpMinus = a - b

--Idee: Liste von links nach rechts auf Stapel legen, wenn CVO Operand, zwei oberste Elem
--poppen, operation durchführen und Ergebnis pushen
eval2 :: Expr -> State2 -> Integer
eval2 ex s = evalStack ex s empty

operator :: CVO -> Bool
operator e = (e == OpPlus) || (e == OpMinus) || (e == OpTimes)

isConstant :: CVO -> Bool
isConstant (Cop c) = True
isConstant _ = False

toChar :: CVO -> Char
toChar (Vop e) = e

evalStack :: Expr -> State2 -> Stack Integer -> Integer
evalStack [] s st = (top1 st)
evalStack (e:es) s st
 | operator e = evalStack es s ((push  (operating (top1 st) (top1 (pop st)) e)) (pop st))
 | isConstant e = evalStack es s (push (constant e) st)
 | otherwise = evalStack es s (push (num2int((s (toChar e)))) st)
 | otherwise = evalStack es s (push (num2int((s (toChar e)))) st)
